#include "GRCC.h"
#include <algorithm>
#include <stdlib.h> // for system()
#include "GRCPrinter.h"
#include "GRCNames.h"
#include "GRCGenOps.h"
#include "GRCMPCGen.h"
#include "Uml.h"
#include "GReATUtils.h"
#include "windows.h"	// FILETIME
#include "GR.h"

///////////////////////////////////////////////////////////////////////////////
// local routines
FILETIME getLastWriteTime( const std::string& filename) {
	HANDLE hFile;
	FILETIME ftCreate, ftLastAccess, ftLastWrite;
	ftLastWrite.dwLowDateTime= 0;
	ftLastWrite.dwHighDateTime= 0;
	hFile = CreateFile( filename.c_str(), GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
	if( hFile != INVALID_HANDLE_VALUE) {	
		GetFileTime( hFile, &ftCreate, &ftLastAccess, &ftLastWrite );
		CloseHandle( hFile);
	}
	return ftLastWrite;
}

bool isGenerationNecessary( const std::string& file, const std::string& dependOnThis) {
	FILETIME fileLastWrite= getLastWriteTime( file);
	FILETIME dependOnLastWrite= getLastWriteTime( dependOnThis);
	if(CompareFileTime(&fileLastWrite, &dependOnLastWrite) == 1) {	
		// fileLastWrite > dependOnLastWrite
		return true;
	}
	return false;
}

///////////////////////////////////////////////////////////////////////////////
// GRCC methods
GRCC::GRCC( const CCodeGenConfig& ccfg) 
	: _ccfg( ccfg), _transformation( GR::diagram)  
{
	_transformation.OpenExisting( ccfg._transFileName, "GR.xsd", Udm::CHANGES_PERSIST_ALWAYS);
}

GRCC::~GRCC() {
	_transformation.CloseNoUpdate();
}

void GRCC::generateLibrary() {
	std::string fnamebase = dropPath( dropExtension( _ccfg._depFileName));
	std::string greatTempDir= getGreatTempDir();
	bool genLib= false;
	if ( _ccfg._dynamic) {
		std::string dll_name = GRCNames::generateDllName( greatTempDir + fnamebase, _ccfg._debug);
		genLib= isGenerationNecessary( _ccfg._depFileName, dll_name);
	} else {
		bool genCode= isGenerationNecessary( _ccfg._depFileName, GRCNames::generateCodeLibName( greatTempDir + fnamebase, _ccfg._debug));
		bool genMeta= isGenerationNecessary( _ccfg._depFileName, GRCNames::generateMetaLibName( greatTempDir + fnamebase, _ccfg._debug));
		bool genOrder= isGenerationNecessary( _ccfg._depFileName, GRCNames::generateOrderLibName( greatTempDir + fnamebase, _ccfg._debug));
		genLib= genCode || genMeta || genOrder;
	}
	//
	if( genLib) {
		std::string restoreDir= getCwd();
		changeDir( greatTempDir);
		// generate AM & Guard code
		generateSourceCode( fnamebase);
		// generate compare template function specialization code
		generateSpecCmpCode( fnamebase);
		// generate .mpc file
		generateMpc( fnamebase);
		// build
		int ret= makeLibrary( fnamebase);
		changeDir( restoreDir);
	}
}

void GRCC::generateSourceCode( const std::string& filename) {
	GR::RootFolder rf = GR::RootFolder::Cast( _transformation.GetRootObject());
	const StrList_t& metanames= _ccfg._metas;
	GRCPrinter grcp( filename, metanames);
	rf.Accept( grcp);
}

void GRCC::generateSpecCmpCode( const std::string& filename) {
	createIncludeMetaFiles( filename, _ccfg._metas);
	GR::RootFolder rf = GR::RootFolder::Cast( _transformation.GetRootObject());
	GRCCmpFuncCollector cmpFuncCollector;
	cmpFuncCollector.Visit_RootFolder( rf);
	generateSpecCompareFunctions( cmpFuncCollector.specTypes(), filename);
}

void GRCC::generateMpc( const std::string& filename) {
	const StrList_t& metanames= _ccfg._metas;
	GRCMPCGen* grcmcpgen= 0;
	if ( _ccfg._dynamic)
		grcmcpgen= new GRCMPCDynamicGen;
	else
		grcmcpgen= new GRCMPCStaticGen;
	grcmcpgen->generate( filename, _ccfg._udmDirName, metanames);
	delete grcmcpgen;
}

static DWORD system2(const char* call) {
	STARTUPINFO si;
    PROCESS_INFORMATION pi;

    ZeroMemory( &si, sizeof(si) );
    si.cb = sizeof(si);
    ZeroMemory( &pi, sizeof(pi) );

	// FIXME: KMS: we'd like to redirect stderr and stdout, but cmd.exe exits with status 1 if the GREngine is
	// invoked via eg python (i.e. command line) with stdout redirected (e.g. testsamples.py > logfile)
	si.hStdOutput = GetStdHandle(STD_OUTPUT_HANDLE);
	si.hStdError = GetStdHandle(STD_OUTPUT_HANDLE);
	//si.dwFlags = STARTF_USESTDHANDLES;

    // Start the child process. 
    if( !CreateProcess( NULL,
        (char*)call,	// Command line
        NULL,           // Process handle not inheritable
        NULL,           // Thread handle not inheritable
        FALSE,          // no handle inheritance
		CREATE_NEW_CONSOLE,
        NULL,           // Use parent's environment block
        NULL,           // Use parent's starting directory 
        &si,
        &pi )
    ) 
    {
		TCHAR cTemp[256];
		DWORD dTemp = GetLastError();
		FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, 0, dTemp, 0, cTemp, 256, NULL);
		throw udm_exception(std::string("CreateProcess(") + call + ") failed: " + cTemp);
    }

    WaitForSingleObject( pi.hProcess, INFINITE );
	DWORD ret = 12;
	GetExitCodeProcess(pi.hProcess, &ret);

    CloseHandle( pi.hProcess );
    CloseHandle( pi.hThread );

	return ret;
}

int GRCC::makeLibrary( const std::string& fnamebase) {
	int sysret;
	cout << endl << "Creating and compiling GRC library." << endl;
	if ( _ccfg._dynamic) {
		std::string compileMode = _ccfg._debug ? "\"Win32 Debug\"" : "\"Win32 Release\"";
		std::string cmdline = "cmd.exe /C call \"%GREAT_PATH%\\etc\\GRC_make.bat\" ";
		cmdline+= "\"" + fnamebase + ".mpc\"" + " " + "\"Makefile." + convertToMpcName( fnamebase) + ".mak\"" + " " + compileMode;
		sysret = system2( cmdline.c_str());
	} else {
		std::string compileMode = _ccfg._debug ? "\"Win32 Static Debug\"" : "\"Win32 Static Release\"";
		std::string cmd = "cmd.exe /C call \"%GREAT_PATH%\\etc\\GRC_make.bat\" ";
		const std::string suffices[]= { "_code", "_meta", "_order"};
		for( int i=0; i<3; ++i) {
			std::string cmdline= cmd+ "\"" + fnamebase + suffices[ i]+ ".mpc\"" + " " + "\"Makefile." + convertToMpcName( fnamebase + suffices[ i])+ ".mak\"" + " " + compileMode;
			sysret = system2( cmdline.c_str());
			if ( sysret)
				break;
		}
	}
	if( sysret == 11) {	
			throw udm_exception("Error: Make process of Dll failed. Unable to find vsvars32.bat via $VS90COMNTOOLS. Please ensure $VS90COMNTOOLS points to a valid location, e.g. c:\\Program Files\\Microsoft Visual Studio 9.0\\Common7\\Tools\\.");
	} else if ( sysret != 0){	
			throw udm_exception("Error: Make process of Dll failed.");
	}
	return sysret;
}
