/* NAME:	Code generator
 * AUTHOR:	Attila Vizhanyo - viza@isis-server.isis.vanderbilt.edu
 *
 * This is the main code generation interface.
 * Usage: user provides a backend consisting of the host meta models, then
 * calls 'generate()' with a GR root folder, then either
 * 1, Get the generated program by 'getProgram()', and print the program code
 *    by calling 'operator<<()'/'printHeader()'/'printSource()' on the program; or
 * -OR-
 * 2, Call 'save()' which saves the code into separate .h,.cpp files plus
 *    print the #include-s for able to compile.
 */
#ifndef __CODEGENERATOR_H
#define __CODEGENERATOR_H

#ifdef _MSC_VER
#pragma warning (disable : 4786)
#pragma warning (disable : 4503) 
#endif
#include <vector>
#include <string>

#include "Sequencer.h"
#include "PE.h"
#include "GeneralBackend.h"
#include "CIntCompiler.h"
#include "CIntLibraryCall.h"

class ConfigGenerator;
///////////////////////////////////////////////////////////////////////////////
class TempRootDefAdder
{
public:
	TempRootDefAdder( Program& program) : _program( program) {}
	void operator()( const std::pair< std::string, GlobalDefinition*>& tempRoot) {
		((PEComposite*)&_program)->add( *tempRoot.second);
	}
private:
	Program& _program;
};

///////////////////////////////////////////////////////////////////////////////
class CodeGenerator
{
public:
	// ctor; Provide all the host meta models, on which the pattern matcher or the effector will execute!
	CodeGenerator( const GeneralBackend& backend, ICIntCompiler& icc, const std::string& gzInc = "") : 
		_backend( backend),
		_program(),
		_passPacket(),
		_translator( _backend, _passPacket, _program, icc) , 
		_sequencer( _translator) ,
		_gzInc( gzInc)
	  {
			GeneralBackend::TempRoots_t tempRoots;
			_backend.getTempRoots( tempRoots);
			std::for_each( tempRoots.begin(), tempRoots.end(), TempRootDefAdder( _program));
	  }
	// dtor
	virtual ~CodeGenerator()
	{}
	// Generates C++ code from a GR model.
	void generate( const GR::RootFolder& grRoot);
	// Generates C++ code from 'rule'.
	void generateRule( const GR::RewritingRule& rule);
	// Caller can gain access to the generated program, to perform printing, etc...
	const Program& getProgram() const { return _program;}
	// Saves the code 'name.cpp' and 'name.h' generated by 'generate()'
	void save( const std::string& name);
	// Save, but also includes 'incFiles' in 'name.cpp''
	void save( const std::string& name, const std::vector< std::string>& incFiles);

private:
	// General backend provides host meta information.
	const GeneralBackend& _backend;
	// Result of code generation.
	Program _program;

protected:
	friend class ConfigGenerator;
	// Does the packet passing between rules. (See PacketPassing.h)
	PacketPassing _passPacket;
	// Translates GE data struct to PE data struct. (See Translator.h)
	Translator _translator;
	// Sequence the rules. (See Sequencer.h)
	Sequencer _sequencer;
	// generated am and guard function header file name
	std::string _gzInc;
};

#endif //__CODEGENERATOR_H