/* NAME:	Pattern matcher
 * AUTHOR:	Attila Vizhanyo - viza@isis-server.isis.vanderbilt.edu
 *
 * Translates patterns of GR rules to C++ code.
 */
#ifndef __PATTERNMATCHER_H
#define __PATTERNMATCHER_H

#ifdef _MSC_VER
#pragma warning (disable : 4786)
#pragma warning (disable : 4503) 
#endif

#include "RuleTypes.h"
#include "PE.h"
#include "GeneralBackend.h"
#include "GReATUtils.h"	 // for 'triplet'
#include "ICIntCompiler.h"

// This struct represents the information of a bound object wrapper  
// needed for doing the pattern matching.
struct BoundObjectInfo
{
	// Definitions which already refer to this bound object wrapper.
	std::list< const Definition*> _references;
	// Definitions of children of this bound object wrapper.
	std::list< const Definition*> _children;
	// Rule names and definitions of object wrappers associated with this bound object wrapper.
	typedef std::pair< std::string, const Definition*> SimpleAssoc_t;
	std::list< SimpleAssoc_t> _simpleAssociations;
	// Rule names and definitions of association classes and object wrappers associated with this bound object wrapper.
	typedef triplet< std::string, const Definition*, const Definition*> AssocClass_t;
	std::list< AssocClass_t> _classAssociations;
};

// 'BoundInfo' represents a collection of 'BoundObjectInfo'-s, 
// and the associated handlers of that collection.
class BoundInfo
{
public:
	void addObjectReference( const GR::ObjectWrapper& ow, const Definition& referDef);
	void addObjectChild( const GR::ObjectWrapper& ow, const Definition& childDef);
	void addObjectSimpleAssociation( const GR::ObjectWrapper& ow, const BoundObjectInfo::SimpleAssoc_t& assocDef);
	void addObjectAssocClass( const GR::ObjectWrapper& ow, const BoundObjectInfo::AssocClass_t& assocDef);
	const BoundObjectInfo* getObjectInfo( const GR::ObjectWrapper& ow) const;
	const std::list< const Definition*>* getAllObjectReferences( const GR::ObjectWrapper& ow) const;
	const std::list< const Definition*>* getAllObjectChildren( const GR::ObjectWrapper& ow) const;
	const std::list< BoundObjectInfo::SimpleAssoc_t>* getAllObjectSimpleAssociations( const GR::ObjectWrapper& ow) const;
	const std::list< BoundObjectInfo::AssocClass_t>* getAllObjectClassAssociations( const GR::ObjectWrapper& ow) const;
	const std::string& getBoundObjectName( const GR::ObjectWrapper& ow) const;
	const Definition* getBoundObjectDefinition( const GR::ObjectWrapper& ow) const; // returns NULL, if object is not bound.

private:
	// Maps object wrappers (representing vertices in a graph pattern) to 'BoundObjectInfo'-s.
	typedef std::map< GR::ObjectWrapper, BoundObjectInfo> BoundObjectInfos_t;
	BoundObjectInfos_t _boundObjects;
};

// Does the pattern matching. 
// Usage: after feeding the proper params in the ctor, call 'createInputMatch()' for every input, 
// then call 'findMatch()' to generate pattern matching code, 
// at last call 'makeMatch()' to generate code filling up the match struct, and returns with the bound objects.
// This class is not designed to be a base class. 
// Either don't inherit from it, or add virtual destructor.
class PatternMatcher
{
public:
	// ctor
	PatternMatcher( const GRSimpleRule& rule, const GeneralBackend& backend, PEComposite& code, ICIntCompiler& cc) 
		:	_rule( rule) , 
			_backend( backend), 
			_currCode( &code) ,
			_cc( cc)
	{}
	// Returns pointer to the type-casted definition, created for 'input'.
	const Definition* createInputMatch( const GR::ObjectWrapper& input, const std::string& inputName);
	// Generate pattern matcher code from '_rule' into '_currCode'.
	void findMatch();
	// Calling 'makeMatch()' after calling 'findMatch()' does 3 things:
	// - Create a Match class and add the class definition to 'ruleDef'. 
	//	 Create a list of Match classes, add the definition to 'ruleDef'.
	// - Create a local temporary variable match of type Match, assigns the match variables 
	//   created by 'findMatch()' to its members, push this match to the end of matches.
	// - Fills up 'boundObjects' mapping pattern objects to their match member names.
	void makeMatch( Class& ruleDef, std::map< GR::ObjectWrapper, std::string>& boundObjects);

protected:
	// Bind 'inputOw', and add reference to 'inputDef' in '_boundObjects'.
	void bindInput( const GR::ObjectWrapper& inputOw, const Definition& inputDef);
	// Bind marks 'ow' to be bound, that is an object wrapper inserted into the match.
	void bind( const GR::ObjectWrapper& ow) const;
	// Returns true, if 'ow' is bound, false otherwise.
	bool isBound( const GR::ObjectWrapper& ow) const;
	// Returns true, if 'edge' is to be matched, false otherwise.
	bool isEdgeToMatch( const GR::ObjectLink& edge) const;
	// Search for an edge, which has both of its vertices is bound. 
	// Returns with true if such found (it is 'edge'), false otherwise.
	bool edgeWithBothVerticesBound( GR::ObjectLink& edge) const;
	// Search for an edge, which has exactly one of its vertices is bound. 
	// Returns with true if such found (it is 'edge'), false otherwise.
	bool edgeWithOneVertexBound( GR::ObjectLink& edge) const;
	// Generates code for parent-child composition relationships, where 'parent' is bound.
	// Returns with reference to Definition PE which is created for the pre-unbound 'child'.
	const Definition& parentBoundComposition( const GR::ObjectWrapper& parent, const GR::ObjectWrapper& child);
	// Generates code for parent-child composition relationships, where 'child' is bound.
	// Returns with reference to Definition PE which is created for the pre-unbound 'parent'.
	const Definition& childBoundComposition( const GR::ObjectWrapper& parent, const GR::ObjectWrapper& child);
	// Generates code for parent-child composition relationships, where both 'parent' and 'child' are bound.
	void bothBoundComposition( const GR::ObjectWrapper& parent, const GR::ObjectWrapper& child);
	//
	const Definition& matchAnyAssociation( const GR::ObjectLink& anyAssocLink);
	//
	void bothBoundMatchAnyAssociation( const GR::ObjectLink& anyAssocLink);
	//
	const Definition& createAnyAssociationCommonCode( const GR::ObjectWrapper& bound, const GR::ObjectWrapper& unbound, const GR::ObjectWrapper& assocClass);
	//
	void associationClassAny( const GR::ObjectLink& assocClassLink);
	// Generates code for simple association type links, where either end of the link is unbound, the other is bound.
	// Returns with reference to Definition PE which is created for the pre-unbound end.
	const Definition& simpleAssociation( const GR::ObjectLink& simpleAssocLink);
	// Generates code for simple association type links, where both ends of the link is bound.
	void bothBoundSimpleAssociation( const GR::ObjectLink& simpleAssocLink);
	// Generates code for association class type links, where the association class type end of the link is unbound.
	// Returns with reference to Definition PE which is created for the pre-unbound end.
	const Definition& assocClassUnboundAssociation( const GR::ObjectLink& assocClassLink);
	// Generates code for association class type links, where the association class type end of the link is bound.
	// Returns with reference to Definition PE which is created for the pre-unbound end.
	const Definition& assocClassBoundAssociation( const GR::ObjectLink& assocClassLink);
	// Generates code for association class type links, where both ends of the link is bound.
	void bothBoundAssociationClass( const GR::ObjectLink& assocClassLink);
	// Main control function; dispatches the processing of 'edge' with both ends bound, depended on the edge type.
	void translateEdgeWithBothVerticesBound( const GR::ObjectLink& edge);
	// Main control function; dispatches the processing of 'edge' with exactly one end bound, depended on the edge type.
	void translateEdgeWithOneVertexBound( const GR::ObjectLink& edge);
	// Validation means checking for equality of different variables, if all of them refer to the same object in the match.
	// Here, code generated which checks 'toCompare' against 'references' for equality.
	void validateReferences( const Definition& toCompare, const std::list< const Definition*>& references);
	// Identity checkers generate code which ensure that all objects in the match are mutually different.
	// This is a neccessary task to do, if-only-if, objects of a match are of the same type, 
	// and connected to some identical object (such as a parent, or other end of an association).
	void checkChildrenIdentities( const Definition& newChild, const std::list< const Definition*>& adjacentChildren);
	void checkSimpleAssocsIdentities( const BoundObjectInfo::SimpleAssoc_t& newAssoc, const std::list< BoundObjectInfo::SimpleAssoc_t>& adjacentAssocs);
	void checkClassAssocsIdentities( const BoundObjectInfo::AssocClass_t& newClassAssoc, const std::list< BoundObjectInfo::AssocClass_t>& adjacentAssocs);
	// get the object wrapper connected to 'assocClass' by 'AssociationClass' type link.
	GR::ObjectWrapper otherEndOfAssociationClass( const GR::ObjectWrapper& assocClass, bool getSources);
	// Used for the case when association class type links in the meta are specified as simple associations in the pattern.
	// In these cases, code must be generated which incorporates the use of association classes.
	// This function creates an object wrapper under 'simpleLink.parent()', which refers to an association class
	// with paradigm 'paradigmName', type 'classType'. Also creates a simple association connecting the association class and 
	// the object wrapper on the proper end of 'simpleLink'. This association class is then returned in 'assocClass'.
	void createPseudoAssocClass( const GR::ObjectLink& simpleLink, const std::string& paradigmName, const string& classType, GR::ObjectWrapper& assocClass) const;
	// Used for the case when association class type links in the meta are specified as simple associations in the pattern.
	// In these cases, code must be generated which incorporates the use of association classes.
	// This function creates an object link under 'simpleLink.parent()', connecting the association class 'assocClass' and 
	// the object wrapper on the proper end of 'simpleAssocLink'. Then executes 'assocClassBoundAssociation()' pattern matching
	// on the created link, so the return value is the reference to the definition generated for the unbound non-association class type end.
	const Definition& createPseudoAssocLink( const GR::ObjectLink& simpleAssocLink, const GR::ObjectWrapper& assocClass);
	// Returns true if either 'type1' or 'type2' is a (possibly indirect) descendant of the other.
	// Types must be of the format: <paradigm>::<type>.
	bool sameTypes( const std::string& type1, const std::string& type2) const;
	// Edge traverse maintenance functions.
	bool isEdgeTraversed( const GR::ObjectLink& edge) const;
	void setEdgeTraversed( const GR::ObjectLink& edge);
	//
	const std::string& getBoundObjectName( const GR::ObjectWrapper& ow);

private:
	// The rule on which pattern matching executes.
	GRSimpleRule _rule;
	// pointer to program element, which is the current target of pattern matching code generation.
	PEComposite* _currCode;
	// Keeps track of edge traversing.
	typedef std::set< GR::ObjectLink> EdgeSet_t;
	EdgeSet_t _traversedEdges;
	// Saves necessary informations for pattern matching. (See class BoundInfo and BoundObjectInfo for more information.)
	BoundInfo _boundObjects;
	// General backend providing host meta information.
	const GeneralBackend& _backend;
	//
	ICIntCompiler& _cc;
};

#endif //__PATTERNMATCHER_H