/* NAME:	Smart DN Code Printer
 * AUTHOR:	Attila Vizhanyo - viza@isis.vanderbilt.edu
 *
 * Code Printer classes for smart data networks
 */
#ifndef SMARTDNCODEPRINTER_H
#define SMARTDNCODEPRINTER_H

#include "DNCodePrinter.h"

class SmartDNCodePrinter : public DNCodePrinter
{
public:
	SmartDNCodePrinter();
	virtual void generate( 
			const std::string& hostFileName,
			const std::string& copyFileName,
			const std::string& meta, 
			const std::string& paradigm, 
			const std::string& root,
			const std::string& dtd
		)
	{
		generateOpen( hostFileName, meta, paradigm, root, dtd);
		generateClose( copyFileName, meta, paradigm, root, dtd);
	}
	virtual void generateOpen( const std::string& fname, const std::string& meta, const std::string& paradigm, const std::string& root, const std::string& dtd) = 0;
	virtual void generateClose( const std::string& fname, const std::string& meta, const std::string& paradigm, const std::string& root, const std::string& dtd) = 0;
	virtual void getOpenCode( PEComposite& addTo) const;
	virtual void getCloseCode( PEComposite& addTo) const;
	virtual void getDnCode( PEComposite& addTo) const;
	virtual const std::string& getDN() const
	{
		return _hostDef->_variable;
	}

protected:
	void createOpenCode( const std::string& fname, const std::string& meta, const std::string& paradigm, const std::string& root, const std::string& dtd);
	void createWriteCode( const std::string& fname, const std::string& meta, const std::string& paradigm, const std::string& root, const std::string& dtd);
	void createOpenCodeWriteOnly( const std::string& meta, const std::string& root);
	void createCopyMemToHostDN();
	void createCopyMemToDN( const std::string& dn);
	void createCloseDNs( bool closeHostWithUpdate, bool closeMemWithUpdate);
	void createCloseHost( bool closeHostWithUpdate);
	void createCloseMem( bool closeMemWithUpdate);
	bool isDTDMeta( const std::string& fname) const;
	
	Definition* _hostDef;
	Expression* _hostRTTExpr;
private:
	Expression* _hostOpenExpr;
	Expression* _hostWriteExpr;
//	Definition* _memDef;
//	Expression* _memCreateExpr;
//	Expression* _copyDNExpr;
//	Expression* _copyBackDNExpr;
	Expression* _hostClose;
//	Expression* _memClose;
};

class SmartDNReadOnlyCodePrinter : public SmartDNCodePrinter
{
public:
	virtual void generateOpen( const std::string& fname, const std::string& meta, const std::string& paradigm, const std::string& root, const std::string& dtd)
	{
		createOpenCode( fname, meta, paradigm, root, dtd);
	}
	virtual void generateClose( const std::string& fname, const std::string& meta, const std::string& paradigm, const std::string& root, const std::string& dtd)
	{
		createCloseDNs( false, false);
	}
	virtual void getCloseCode( PEComposite& addTo) const
	{
		SmartDNCodePrinter::getCloseCode( addTo);
	}
};

class SmartDNWriteOnlyCodePrinter : public SmartDNCodePrinter
{
public:
	SmartDNWriteOnlyCodePrinter();
	virtual void generateOpen( const std::string& fname, const std::string& meta, const std::string& paradigm, const std::string& root, const std::string& dtd)
	{
//		createOpenCodeWriteOnly( meta, root);
		createWriteCode( fname, meta, paradigm, root, dtd);
	}
	virtual void generateClose( const std::string& fname, const std::string& meta, const std::string& paradigm, const std::string& root, const std::string& dtd)
	{
		createCloseDNs( true, false);
//		createNewDN( fname, meta, paradigm, root, dtd);
//		createCopyMemToNewDN();
//		createCloseMem( true);
//		createCloseNewDN();
	}
	virtual void getCloseCode( PEComposite& addTo) const
	{
//		if ( _newHostDef) addTo.add( *_newHostDef);
//		if ( _createNewHostExpr) addTo.add( *_createNewHostExpr);
		SmartDNCodePrinter::getCloseCode( addTo);
//		if ( _closeNewHostExpr) addTo.add( *_closeNewHostExpr);
	}

protected:
	virtual const std::string getNewHostPrefix() 
	{
		return "sdnNew";
	}
	void createNewDN( const std::string& fname, const std::string& meta, const std::string& paradigm, const std::string& root, const std::string& dtd);
	void createCopyMemToNewDN();
	void createCloseNewDN();

private:
	Definition* _newHostDef;
	Expression* _createNewHostExpr;
	Expression* _closeNewHostExpr;
};

class SmartDNReadWriteCodePrinter : public SmartDNCodePrinter
{
public:
	virtual void generateOpen( const std::string& fname, const std::string& meta, const std::string& paradigm, const std::string& root, const std::string& dtd)
	{
		createOpenCode( fname, meta, paradigm, root, dtd);
	}
	virtual void generateClose( const std::string& fname, const std::string& meta, const std::string& paradigm, const std::string& root, const std::string& dtd)
	{
//		createCopyMemToHostDN();
		createCloseDNs( true, false);
	}
	virtual void getCloseCode( PEComposite& addTo) const
	{
		SmartDNCodePrinter::getCloseCode( addTo);
	}
};

class SmartDNWriteCopyCodePrinter : public SmartDNWriteOnlyCodePrinter
{
public:
	SmartDNWriteCopyCodePrinter() : SmartDNWriteOnlyCodePrinter() {}
	virtual void generateOpen( const std::string& fname, const std::string& meta, const std::string& paradigm, const std::string& root, const std::string& dtd)
	{
		createOpenCode( fname, meta, paradigm, root, dtd);
	}
	virtual void generateClose( const std::string& fname, const std::string& meta, const std::string& paradigm, const std::string& root, const std::string& dtd);
	virtual void getCloseCode( PEComposite& addTo) const
	{
		addTo.add(*new Expression(_hostDef->_variable+".SaveAs("+_copyname+")"));
		SmartDNWriteOnlyCodePrinter::getCloseCode( addTo);
	}

protected:
	virtual const std::string getNewHostPrefix()
	{
		return "sdnCopy";
	}
private:
	std::string _copyname;
};

#endif SMARTDNCODEPRINTER_H