///////////////////////////////////////////////////////////////////////////
// RawComponent.cpp, the main RAW COM component implementation file
// This is the file (along with its header RawComponent.h)
// that the component implementor is expected to modify in the first place
//
///////////////////////////////////////////////////////////////////////////
#include "stdafx.h"

#include "Console.h"
#include "ComHelp.h"
#include "GMECOM.h"
#include "ComponentConfig.h"
#include "UdmConfig.h"
#include "RawComponent.h"

// Udm includes
#include "UdmBase.h"
#include "Uml.h"
#include "UmlExt.h"

#ifdef _USE_DOM
	#include "UdmDOM.h"
#endif

#include "UdmGme.h"
#include "UdmStatic.h"
#include "UdmUtil.h"

#include "UdmApp.h"

#include "umt_exception.h"

#include <iostream>
#include <fstream>

#define GREAT_SILENT_MODE 128

// Global config object
_config config;


// this method is called after all the generic initialization is done
// this should be empty, unless application-specific initialization is needed
STDMETHODIMP RawComponent::Initialize(struct IMgaProject *) {
	return S_OK;
}

// this is the obsolete component interface
// this present implementation either tries to call InvokeEx, or returns an error;
STDMETHODIMP RawComponent::Invoke(IMgaProject* gme, IMgaFCOs *models, long param) {
#ifdef SUPPORT_OLD_INVOKE
	CComPtr<IMgaFCO> focus;
	CComVariant parval = param;
	return InvokeEx(gme, focus, selected, parvar);
#else
	if(interactive) {
		AfxMessageBox("This component does not support the obsolete invoke mechanism");
	}
	return E_MGA_NOT_SUPPORTED;
#endif
}

#ifdef _DYNAMIC_META
			void dummy(void) {; } // Dummy function for UDM meta initialization
#endif


void appendErrorLog(std::string &tmp_str)
{
	std::ifstream in((CUdmApp::mgaDir+ "errorlog.txt").c_str());
	std::string line;
	while(getline(in, line))
		tmp_str += line + "\n";
}

void closeErrorLog(std::string &tmp_str)
{
	if(tmp_str.empty()) return;
	std::ofstream errLogFile;
	errLogFile.open((CUdmApp::mgaDir+ "errorlog.txt").c_str());
	errLogFile << tmp_str;
	errLogFile.close();
}
// This is the main component method for interpereters and plugins. 
// May als be used in case of invokeable addons
STDMETHODIMP RawComponent::InvokeEx( IMgaProject *project,  IMgaFCO *currentobj,  
									IMgaFCOs *selectedobjs,  long param) 
{
	// Calling the user's initialization function
	if(CUdmApp::Initialize())
	{
		return S_FALSE;
	}
	if (param == GREAT_SILENT_MODE || !interactive) {
		interactive = false;
		param = GREAT_SILENT_MODE;
	}

	CComPtr<IMgaProject>ccpProject(project);
	try
	{	  
		CComBSTR projname;
		CComBSTR focusname = "<nothing>";

		CComPtr<IMgaProject> mgaProject = project;
		CComBSTR    connString;
		COMTHROW(mgaProject->get_ProjectConnStr(&connString));

		CString mgaFilePath = connString;
		mgaFilePath.Replace('/', '\\');
		int e = mgaFilePath.ReverseFind('\\');
		if(e != -1)
		{	
			CUdmApp::mgaDir = mgaFilePath.Mid(4, e-3);
		}


		// Setting up Udm	
#ifdef _DYNAMIC_META
	#ifdef _DYNAMIC_META_DOM
			// Loading the meta for the project
			UdmDom::DomDataNetwork  ddnMeta(Uml::diagram);	
			Uml::Diagram theUmlDiagram;

			// Opening the XML meta of the project
			ddnMeta.OpenExisting(config.metaPath,"uml.xsd", Udm::CHANGES_LOST_DEFAULT);
			
			// Casting the DataNetwork to diagram
			theUmlDiagram = Uml::Diagram::Cast(ddnMeta.GetRootObject());
			
			// Creating the UDM diagram
			Udm::UdmDiagram udmDataDiagram;
			udmDataDiagram.dgr = &theUmlDiagram;
			udmDataDiagram.init = dummy;
		
	#elif defined _DYNAMIC_META_STATIC
			// Loading the meta for the project
			UdmStatic::StaticDataNetwork  dnsMeta(Uml::diagram);	
			Uml::Diagram theUmlDiagram;

			// Opening the static meta of the project
			dnsMeta.OpenExisting(config.metaPath, "", Udm::CHANGES_LOST_DEFAULT);
			
			// Casting the DataNetwork to diagram
			theUmlDiagram = Uml::Diagram::Cast(dnsMeta.GetRootObject());
			
			// Creating the UDM diagram
			Udm::UdmDiagram udmDataDiagram;
			udmDataDiagram.dgr = &theUmlDiagram;
			udmDataDiagram.init = dummy;

	#else
			ASSERT((0,"Nor _DYNAMIC_META_DOM either _DYNAMIC_META_STATIC defined for dynamic loading"));
	#endif
			// Loading the project
			UdmGme::GmeDataNetwork dngBackend(udmDataDiagram);		
		
#else
		using namespace META_NAMESPACE;
		
		// Loading the project			
		UdmGme::GmeDataNetwork dngBackend(META_NAMESPACE::diagram);				
					
#endif
		try
		{
			// Opening backend
			dngBackend.OpenExisting(ccpProject, Udm::CHANGES_LOST_DEFAULT);
			

			CComPtr<IMgaFCO> ccpFocus(currentobj);
			Udm::Object currentObject;
			if(ccpFocus)
			{
				currentObject=dngBackend.Gme2Udm(ccpFocus);
			}

			set<Udm::Object> selectedObjects;

			CComPtr<IMgaFCOs> ccpSelObject(selectedobjs);

			MGACOLL_ITERATE(IMgaFCO,ccpSelObject){		 
				Udm::Object currObj;
				if(MGACOLL_ITER)
				{
					currObj=dngBackend.Gme2Udm(MGACOLL_ITER);
				}
			 selectedObjects.insert(currObj);
			}MGACOLL_ITERATE_END;

#ifdef _ACCESS_MEMORY
			// Creating Cache
	#ifdef _DYNAMIC_META
			UdmStatic::StaticDataNetwork dnsCacheBackend(udmDataDiagram);
	#else
			UdmStatic::StaticDataNetwork dnsCacheBackend(META_NAMESPACE::diagram);
	#endif
			
			const Uml::Class & safeType = Uml::SafeTypeContainer::GetSafeType(dngBackend.GetRootObject().type());

			dnsCacheBackend.CreateNew("","",safeType, Udm::CHANGES_LOST_DEFAULT); 
			
			Udm::Object nullObject(&Udm::__null);
			UdmUtil::copy_assoc_map copyAssocMap;
			copyAssocMap[currentObject]=nullObject; // currentObject may be null object
		
			for(set<Udm::Object>::iterator p_CurrSelObject=selectedObjects.begin();
				p_CurrSelObject!=selectedObjects.end();p_CurrSelObject++)
			{
					pair<Udm::Object const, Udm::Object> item(*p_CurrSelObject, nullObject);

					pair<UdmUtil::copy_assoc_map::iterator, bool> insRes = copyAssocMap.insert(item);

					if (!insRes.second)
					{
						ASSERT(NULL);
					}

			}

			// Copying from GME to memory
			UdmUtil::CopyObjectHierarchy(
				dngBackend.GetRootObject().__impl(),
				dnsCacheBackend.GetRootObject().__impl(),
				&dnsCacheBackend,
				copyAssocMap);
		
			// Searching for focus object
			Udm::Object currentObjectCache;
			UdmUtil::copy_assoc_map::iterator currObject = copyAssocMap.find(currentObject);
			if (currObject != copyAssocMap.end()) // It is in the map
			{
				currentObjectCache=currObject->second;
			}


			// Searching for selected objects
			set<Udm::Object> selectedObjectsCache;

			for( p_CurrSelObject=selectedObjects.begin();
				p_CurrSelObject!=selectedObjects.end();p_CurrSelObject++)
			{
				Udm::Object object;
				UdmUtil::copy_assoc_map::iterator currSelObjectIt = copyAssocMap.find(*p_CurrSelObject);
				if (currSelObjectIt != copyAssocMap.end()) // It is in the map
				{
					object=currSelObjectIt->second;
					selectedObjectsCache.insert(object);
				}	
			}
	

			// Closing GME backend
			dngBackend.CloseNoUpdate();

			// Calling the main entry point
			CUdmApp::UdmMain(&dnsCacheBackend,currentObjectCache,selectedObjectsCache,param);
			// Close cache backend
			dnsCacheBackend.CloseNoUpdate();

#else
			// Calling the main entry point
			CUdmApp::UdmMain(&dngBackend,currentObject,selectedObjects,param);
			// Closing backend
			dngBackend.CloseWithUpdate();
			
#endif

			// Code for the interpterer running sequence comes here

			std::string tmp_str;
			try 
			{	
				CComPtr<IMgaProject> spProject = project;
				CComPtr<IMgaFCO> spContextFCO = currentobj;
				CComPtr<IMgaFCOs> spSelectedFCOs = selectedobjs;
				long lParam = param;
				VARIANT_BOOL vbInteractive = interactive ? VARIANT_TRUE : VARIANT_FALSE;

				if(CUdmApp::genGR == TRUE)
				{	// Calling the GR interprerter
					CComBSTR bstrCompName( L"MGA.Interpreter.Generate GR" );

					CComPtr<IMgaComponentEx> grComponent;
					COMTHROW( grComponent.CoCreateInstance( bstrCompName ) ) ;
					COMTHROW( grComponent->put_InteractiveMode( vbInteractive ) );
					COMTHROW( grComponent->Enable( VARIANT_TRUE ) );
					COMTHROW( grComponent->Initialize( spProject ) );
					HRESULT ret = grComponent->InvokeEx( spProject, spContextFCO, spSelectedFCOs, lParam ) ;
					appendErrorLog(tmp_str);
					if(ret != S_OK)
					{
						closeErrorLog(tmp_str);	
						return ret;
					}				
				}
				
				if(CUdmApp::genGR == TRUE || CUdmApp::genCFG == TRUE || CUdmApp::genUDM == TRUE)
				{	// Calling the GRC interprerter
					CComBSTR bstrCompName( L"MGA.Interpreter.Generate GRC" );

					CComPtr<IMgaComponentEx> grComponent;
					COMTHROW( grComponent.CoCreateInstance( bstrCompName ) ) ;
					COMTHROW( grComponent->put_InteractiveMode( vbInteractive ) );
					COMTHROW( grComponent->Enable( VARIANT_TRUE ) );
					COMTHROW( grComponent->Initialize( spProject ) );
					HRESULT ret = grComponent->InvokeEx( spProject, spContextFCO, spSelectedFCOs, lParam ) ;
					if(ret != S_OK)
					{
						closeErrorLog(tmp_str);	
						return ret;
					}
				}

				if(CUdmApp::genUDM == TRUE)
				{	// Calling the UML2XML interprerter
					CComBSTR bstrCompName( L"MGA.Interpreter.UML2XML" );
					
					CComPtr<IMgaComponentEx> grComponent;
					COMTHROW( grComponent.CoCreateInstance( bstrCompName ) ) ;
					COMTHROW( grComponent->put_InteractiveMode( vbInteractive ) );
					COMTHROW( grComponent->Enable( VARIANT_TRUE ) );
					COMTHROW( grComponent->Initialize( spProject ) );
					HRESULT ret = grComponent->InvokeEx( spProject, spContextFCO, spSelectedFCOs, lParam ) ;
					if(ret != S_OK)
					{
						closeErrorLog(tmp_str);						
						return ret;
					}
				}

				if(CUdmApp::genCFG == TRUE)
				{	// Calling the GenerateConfig interprerter
					CComBSTR bstrCompName( L"MGA.Interpreter.GenerateConfig" );
					
					CComPtr<IMgaComponentEx> grComponent;
					COMTHROW( grComponent.CoCreateInstance( bstrCompName ) ) ;
					COMTHROW( grComponent->put_InteractiveMode( vbInteractive ) );
					COMTHROW( grComponent->Enable( VARIANT_TRUE ) );
					COMTHROW( grComponent->Initialize( spProject ) );
					HRESULT ret = grComponent->InvokeEx( spProject, spContextFCO, spSelectedFCOs, lParam ) ;
					appendErrorLog(tmp_str);
					if(ret != S_OK)
					{
						closeErrorLog(tmp_str);	
						return ret;
					}				
				}

				
				if(CUdmApp::runGRE == TRUE)
				{	// Calling the Populate File Info interprerter
					CComBSTR bstrCompName( L"MGA.Interpreter.InvokeGRengine" );

					CComPtr<IMgaComponentEx> grComponent;
					COMTHROW( grComponent.CoCreateInstance( bstrCompName ) ) ;
					COMTHROW( grComponent->put_InteractiveMode( VARIANT_TRUE ) );
					COMTHROW( grComponent->Enable( VARIANT_TRUE ) );
					COMTHROW( grComponent->Initialize( spProject ) );
					HRESULT ret = grComponent->InvokeEx( spProject, spContextFCO, spSelectedFCOs, lParam ) ;
					if(ret != S_OK)
					{
						closeErrorLog(tmp_str);
						return ret;
					}
				}

				if(CUdmApp::runCG == TRUE)
				{	// Calling the Populate File Info interprerter
					CComBSTR bstrCompName( L"Generate code from rules" );

					CComPtr<IMgaComponentEx> grComponent;
					COMTHROW( grComponent.CoCreateInstance( bstrCompName ) ) ;
					COMTHROW( grComponent->put_InteractiveMode( VARIANT_TRUE ) );
					COMTHROW( grComponent->Enable( VARIANT_TRUE ) );
					COMTHROW( grComponent->Initialize( spProject ) );
					HRESULT ret = grComponent->InvokeEx( spProject, spContextFCO, spSelectedFCOs, lParam ) ;
					if(ret != S_OK)
					{
						closeErrorLog(tmp_str);
						return ret;
					}
				}	
				closeErrorLog(tmp_str);	
				
				string configDir;
				int ne = CUdmApp::configName.find_last_of('\\');
				configDir = CUdmApp::configName.substr(0,ne+1);

				CString msg = "Files Generated:\n";
				if(CUdmApp::genCFG == TRUE)
					msg = msg + CUdmApp::configName.c_str() + "\n";
				if(CUdmApp::genGR == TRUE)
					msg = msg + configDir.c_str() + CUdmApp::grName.c_str() + "\n";
				if(CUdmApp::genUDM == TRUE)
					msg = msg + configDir.c_str() + CUdmApp::udmName.c_str() + "\n";
				if(CUdmApp::genCFG == TRUE || CUdmApp::genGR == TRUE || CUdmApp::genUDM == TRUE)
					GMEConsole::Console::Info::writeLine(msg);
											
			}	
			catch(...) 
			{ 
				closeErrorLog(tmp_str);
				if(param != 128)
				{
					AfxMessageBox("Unknown Master Interpreter Exception");
				}
				GMEConsole::Console::Error::WriteLine("Unknown Master Interpreter Exception");
				return E_FAIL;
			}
		}
		catch(const umt_exception& exc)
		{	
#ifdef _META_ACCESS_MEMORY
			dnCacheBackend.CloseNoUpdate();
#endif
			// Close GME Backend (we may close it twice, but GmeDataNetwork handles it)
			dngBackend.CloseNoUpdate();
			GMEConsole::Console::Error::WriteLine(exc.what().c_str());
			return E_FAIL;
		}
		catch(udm_exception &exc) 
		{
#ifdef _META_ACCESS_MEMORY
			dnCacheBackend.CloseNoUpdate();
#endif
			// Close GME Backend (we may close it twice, but GmeDataNetwork handles it)
			dngBackend.CloseNoUpdate();			

			if(param != 128)
			{
				AfxMessageBox(exc.what(),MB_ICONSTOP);			
			}
			GMEConsole::Console::Error::WriteLine(exc.what());
			CComPtr<IMgaTerritory> terr;
			ccpProject->get_ActiveTerritory(&terr);
			if (terr) {
				ccpProject->AbortTransaction();
			}
			return E_FAIL;
		}
	}
	catch(...) 
	{
		ccpProject->AbortTransaction();
		if(param != 128)
		{
			AfxMessageBox("An unexpected error has occured during the interpretation process.");
		}
		GMEConsole::Console::Error::WriteLine("An unexpected error has occured during the interpretation process.");
		return E_UNEXPECTED;
	}
	return S_OK;
}

// GME currently does not use this function
// you only need to implement it if other invokation mechanisms are used
STDMETHODIMP RawComponent::ObjectsInvokeEx( IMgaProject *project,  IMgaObject *currentobj,  IMgaObjects *selectedobjs,  long param) {
	if(interactive) {
		AfxMessageBox("Tho ObjectsInvoke method is not implemented");
	}
	return E_MGA_NOT_SUPPORTED;
}


// implement application specific parameter-mechanism in these functions:
STDMETHODIMP RawComponent::get_ComponentParameter(BSTR name, VARIANT *pVal) {
	return S_OK;
}

STDMETHODIMP RawComponent::put_ComponentParameter(BSTR name, VARIANT newVal) {
	return S_OK;
}


#ifdef GME_ADDON

// these two functions are the main 
STDMETHODIMP RawComponent::GlobalEvent(globalevent_enum event) { 
	if(event == GLOBALEVENT_UNDO) {
		AfxMessageBox("UNDO!!");
	}
	return S_OK; 
}

STDMETHODIMP RawComponent::ObjectEvent(IMgaObject * obj, unsigned long eventmask, VARIANT v) {
	if(eventmask & OBJEVENT_CREATED) {
		CComBSTR objID;
		COMTHROW(obj->get_ID(&objID));
		AfxMessageBox( "Object created! ObjID: " + CString(objID)); 
	}		
	return S_OK;
}

#endif
